/*
Ultima 4 World Map Editor v1.1

Copyright 2011 Matthew Duncan

This file is part of the Ultima 4 World Map Editor v1.1.

Ultima 4 World Map Editor v1.1 is free software: you can redistribute it and/or
modify it under the terms of the GNU General Public License as published by the
Free Software Foundation, either version 3 of the License, or (at your option)
any later version.

Ultima 4 World Map Editor v1.1 is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License
along with Ultima 4 World Map Editor v1.1. If not, see
<http://www.gnu.org/licenses/>.
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <allegro.h>
#include "u4maped.1.1.h"

int main(int argc, char **argv)
{
	int Done, KeyStroke, i;
	if (allegro_init() != 0) return 1;
	install_keyboard();
	install_timer();

	if (!InitVars())
	{
		perror("can't init vars");
		return(1);
	}
	if (!LoadTiles())
	{
		perror("can't load tiles");
		return(1);
	}
	if (!LoadMap())
	{
		perror("can't load map");
		return(1);
	}
//	clear(screen);
/*	printf("\n\nAllegro init... OK\n");
	printf("version %s\n", allegro_id);
	printf("\n\nUltima 4 Map Editor v1.1 by Ram Dragon\n");
	printf("\n\nUse arrow keys to move cursor\n");
	printf("Use space bar to toggle maps\nUse letter keys a-z..A-Z to paint tiles onto the map\n");
	printf("Press 'ESC' to quit the editor\n\n");
	printf("press a key to start...");
	clear_keybuf();
	readkey();*/
	if (set_gfx_mode(GFX_AUTODETECT, GModeW, GModeH, 0, 0) != 0)
	{
		if (set_gfx_mode(GFX_SAFE, GModeW, GModeH, 0, 0) != 0)
		{
			set_gfx_mode(GFX_TEXT, 0, 0, 0, 0);
			allegro_message("Unable to set any graphic mode\n%s\n", allegro_error);
			return 1;
		}
	}

	set_pallete(UpgradePalette);
	clear_to_color(screen, makecol(0, 0, 0));

	PaintMaps();
	CursorToggle = ToggleCursor(CursorToggle);
	PaintEditTiles();
	Done = 0;
	while (!Done)
	{
		clear_keybuf();
		KeyStroke = readkey();
		switch((KeyStroke >> 8))
		{
			case KEY_SPACE: CursorToggle = ToggleCursor(CursorToggle); break;
			case KEY_UP: if (CursorToggle) { EraseMapCursor(); MapY = MoveCursor(MapY, UP, MAPMIN, MAPMAX); DrawMapCursor(); ReadMapCursor(); break; }
						else { EraseEditCursor(); EditY = MoveCursor(EditY, UP, EDITMIN, EDITMAX); DrawEditCursor(); break; }
			case KEY_DOWN: if (CursorToggle) { EraseMapCursor(); MapY = MoveCursor(MapY, DOWN, MAPMIN, MAPMAX); DrawMapCursor(); ReadMapCursor(); break; }
						else { EraseEditCursor(); EditY = MoveCursor(EditY, DOWN, EDITMIN, EDITMAX); DrawEditCursor(); break; }
			case KEY_LEFT: if (CursorToggle) { EraseMapCursor(); MapX = MoveCursor(MapX, LEFT, MAPMIN, MAPMAX); DrawMapCursor(); ReadMapCursor(); break; }
						else { EraseEditCursor(); EditX = MoveCursor(EditX, LEFT, EDITMIN, EDITMAX); DrawEditCursor(); break; }
			case KEY_RIGHT: if (CursorToggle) { EraseMapCursor(); MapX = MoveCursor(MapX, RIGHT, MAPMIN, MAPMAX); DrawMapCursor(); ReadMapCursor(); break; }
				        else { EraseEditCursor(); EditX = MoveCursor(EditX, RIGHT, EDITMIN, EDITMAX); DrawEditCursor(); break; }
			case KEY_ESC: Done = 1; break;
			case KEY_1: if (!SaveMap()) { destroy_bitmap(MapCursorPic); for(i=0; i<MAX_TILES; i++) destroy_bitmap(Tiles[i]); return(1);} else break;
			case KEY_2: RestoreMap(); break;
			case KEY_3: save_bmp("u4maped.bmp", screen, UpgradePalette);
			case KEY_Q: if (key_shifts & KB_SHIFT_FLAG) EditScreen(TileNumberArray[26]); else EditScreen(TileNumberArray[0]); break;
			case KEY_W: if (key_shifts & KB_SHIFT_FLAG) EditScreen(TileNumberArray[27]); else EditScreen(TileNumberArray[1]); break;
			case KEY_E: if (key_shifts & KB_SHIFT_FLAG) EditScreen(TileNumberArray[28]); else EditScreen(TileNumberArray[2]); break;
			case KEY_R: if (key_shifts & KB_SHIFT_FLAG) EditScreen(TileNumberArray[29]); else EditScreen(TileNumberArray[3]); break;
			case KEY_T: if (key_shifts & KB_SHIFT_FLAG) EditScreen(TileNumberArray[30]); else EditScreen(TileNumberArray[4]); break;
			case KEY_Y: if (key_shifts & KB_SHIFT_FLAG) EditScreen(TileNumberArray[31]); else EditScreen(TileNumberArray[5]); break;
			case KEY_U: if (key_shifts & KB_SHIFT_FLAG) EditScreen(TileNumberArray[32]); else EditScreen(TileNumberArray[6]); break;
			case KEY_I: if (key_shifts & KB_SHIFT_FLAG) EditScreen(TileNumberArray[33]); else EditScreen(TileNumberArray[7]); break;
			case KEY_O: if (key_shifts & KB_SHIFT_FLAG) EditScreen(TileNumberArray[34]); else EditScreen(TileNumberArray[8]); break;
			case KEY_P: if (key_shifts & KB_SHIFT_FLAG) EditScreen(TileNumberArray[35]); else EditScreen(TileNumberArray[9]); break;
			case KEY_A: if (key_shifts & KB_SHIFT_FLAG) EditScreen(TileNumberArray[36]); else EditScreen(TileNumberArray[10]); break;
			case KEY_S: if (key_shifts & KB_SHIFT_FLAG) EditScreen(TileNumberArray[37]); else EditScreen(TileNumberArray[11]); break;
			case KEY_D: if (key_shifts & KB_SHIFT_FLAG) EditScreen(TileNumberArray[38]); else EditScreen(TileNumberArray[12]); break;
			case KEY_F: if (key_shifts & KB_SHIFT_FLAG) EditScreen(TileNumberArray[39]); else EditScreen(TileNumberArray[13]); break;
			case KEY_G: if (key_shifts & KB_SHIFT_FLAG) EditScreen(TileNumberArray[40]); else EditScreen(TileNumberArray[14]); break;
			case KEY_H: if (key_shifts & KB_SHIFT_FLAG) EditScreen(TileNumberArray[41]); else EditScreen(TileNumberArray[15]); break;
			case KEY_J: if (key_shifts & KB_SHIFT_FLAG) EditScreen(TileNumberArray[42]); else EditScreen(TileNumberArray[16]); break;
			case KEY_K: if (key_shifts & KB_SHIFT_FLAG) EditScreen(TileNumberArray[43]); else EditScreen(TileNumberArray[17]); break;
			case KEY_L: if (key_shifts & KB_SHIFT_FLAG) EditScreen(TileNumberArray[44]); else EditScreen(TileNumberArray[18]); break;
			case KEY_Z: if (key_shifts & KB_SHIFT_FLAG) EditScreen(TileNumberArray[45]); else EditScreen(TileNumberArray[19]); break;
			case KEY_X: if (key_shifts & KB_SHIFT_FLAG) EditScreen(TileNumberArray[46]); else EditScreen(TileNumberArray[20]); break;
			case KEY_C: if (key_shifts & KB_SHIFT_FLAG) EditScreen(TileNumberArray[47]); else EditScreen(TileNumberArray[21]); break;
			case KEY_V: if (key_shifts & KB_SHIFT_FLAG) EditScreen(TileNumberArray[48]); else EditScreen(TileNumberArray[22]); break;
			case KEY_B: if (key_shifts & KB_SHIFT_FLAG) EditScreen(TileNumberArray[49]); else EditScreen(TileNumberArray[23]); break;
			case KEY_N: if (key_shifts & KB_SHIFT_FLAG) EditScreen(TileNumberArray[50]); else EditScreen(TileNumberArray[24]); break;
			case KEY_M: if (key_shifts & KB_SHIFT_FLAG) EditScreen(TileNumberArray[51]); else EditScreen(TileNumberArray[25]); break;
			default: break;
		}
	}
	destroy_bitmap(MapCursorPic);
	for(i=0; i<MAX_TILES; i++) destroy_bitmap(Tiles[i]);
	return(0);
}
END_OF_MAIN()

/****************************************************************************************/
/****************************************************************************************/
int InitVars(void)
{
	if ((Map = (char *) malloc(sizeof(char)*FILESIZE)) == NULL)
	{
		perror("Error allocating for Map File: ");
		return(0);
	}
	if ((MapBuffer = (char *) malloc(sizeof(char)*FILESIZE)) == NULL)
	{
		perror("Error allocating for Map Buffer: ");
		return(0);
	}
	MapCursorPic = create_bitmap(CURSOR_PIC_SIZE, CURSOR_PIC_SIZE);
	MapX = 15;
	MapY = 15;
	EditX = 7;
	EditY = 7;
	CursorToggle = 0;
	EditFlag = 0;
	return(1);
}
/****************************************************************************************/
/****************************************************************************************/
void U4File2Map(char *U4File, char *MapFile)
{
	unsigned long i, j, k, l, U4FileOffSet, MapFileOffSet;
	for (l=0; l<8; l++)
		for (k=0; k<8; k++)
		{
			U4FileOffSet = k*1024+l*8192;
			MapFileOffSet = k*32+l*8192;
			for (j=0; j<32; j++)
				for (i=0; i<32; i++)
					*(MapFile+MapFileOffSet+i+(j*256)) = *(U4File+U4FileOffSet+i+(j*32));
		}
	for (i=0; i<65536; i++) *(U4File+i) = *(MapFile+i);
}
/****************************************************************************************/
/****************************************************************************************/
int LoadMap(void)
{
	FILE *fp;

	if ((fp = fopen("World.map", "rb")) == NULL)
	{
		perror("error open: ");
		return(0);
	}
	if (fread(Map, 1, MAP_FILE_SIZE, fp) != MAP_FILE_SIZE)
	{
		perror("error read: ");
		return(0);
	}
	fclose(fp);
	U4File2Map(Map, MapBuffer);
	return(1);
}
/****************************************************************************************/
/****************************************************************************************/
int LoadTiles(void)
{
	char Name[256];
	int i;

	for (i=0; i<MAX_TILES; i++)
	{
		Tiles[i] = create_bitmap(TILE_SIZE, TILE_SIZE);
		sprintf(Name, "maptiles\\image%03d.bmp", i+1);
		Tiles[i] = load_bmp(Name, UpgradePalette);
	}
	return(1);
}
/****************************************************************************************/
/****************************************************************************************/
void ReadMapCursor(void)
{
	unsigned long CursorBase, i, j;
	CursorBase = (unsigned long)MapX*CURSORSIZE + (unsigned long)MapY*CURSORSIZE*LINELENGTH;
	for (j=0; j<TILESPERLINE; j++)
		for (i=0; i<TILESPERLINE; i++)
			blit(Tiles[(int) *(MapBuffer+CursorBase+i+(j*LINELENGTH))], screen, 0, 0, i*CURSOROFF + EDITBASEX, j*CURSOROFF + EDITBASEY, TILE_SIZE, TILE_SIZE);
}
/****************************************************************************************/
/****************************************************************************************/
int MoveCursor(int CurrPos, int Direction, int MINVal, int MAXVal)
{
	switch(Direction)
	{
		case UP: if ((CurrPos-1) < MINVal) return(MAXVal); else return(CurrPos-1);
		case DOWN: if ((CurrPos+1) > MAXVal) return(MINVal); else return(CurrPos+1);
		case LEFT: if ((CurrPos-1) < MINVal) return(MAXVal); else return(CurrPos-1);
		case RIGHT: if ((CurrPos+1) > MAXVal) return(MINVal); else return(CurrPos+1);
		default: return(CurrPos);
	}
}
/****************************************************************************************/
/****************************************************************************************/
void DrawMapCursor(void)
{
	int X, Y;
	X = MapX*CURSORSIZE + MAPBASEX-1;
	Y = MapY*CURSORSIZE + MAPBASEY-1;
	blit(screen, MapCursorPic, X, Y, 0, 0, CURSOR_PIC_SIZE, CURSOR_PIC_SIZE);
	rect(screen, X, Y, X+CURSOROFF, Y+CURSOROFF, RED);
}
/****************************************************************************************/
/****************************************************************************************/
void EraseMapCursor(void)
{
	int X, Y;
	X = MapX*CURSORSIZE + MAPBASEX-1;
	Y = MapY*CURSORSIZE + MAPBASEY-1;
	blit(MapCursorPic, screen, 0, 0, X, Y, CURSOR_PIC_SIZE, CURSOR_PIC_SIZE);
}
/****************************************************************************************/
/****************************************************************************************/
void DrawEditCursor(void)
{
	int X, Y;
	X = EditX*CURSOROFF + EDITBASEX-1;
	Y = EditY*CURSOROFF + EDITBASEY-1;
	rect(screen, X, Y, X+CURSOROFF, Y+CURSOROFF, RED);
}
/****************************************************************************************/
/****************************************************************************************/
void EraseEditCursor(void)
{
	int X, Y;
	X = EditX*CURSOROFF + EDITBASEX-1;
	Y = EditY*CURSOROFF + EDITBASEY-1;
	rect(screen, X, Y, X+CURSOROFF, Y+CURSOROFF, BLACK);
}
/****************************************************************************************/
/****************************************************************************************/
void DrawMapSelectFrame(int X, int Y, int SIZE, int STAT)
{
	if (STAT) rect(screen, X, Y, X+SIZE, Y+SIZE, RED);
	else rect(screen, X, Y, X+SIZE, Y+SIZE, BLACK);
}
/****************************************************************************************/
/****************************************************************************************/
int ToggleCursor(int CurrStatus)
{
	if (CurrStatus)
	{
		DrawMapSelectFrame(MAPBASEX-2, MAPBASEY-2, MAPSELECTOFF, 0);
		DrawMapSelectFrame(EDITBASEX-2, EDITBASEY-2, EDITSELECTOFF, 1);
		ReadMapCursor();
		return(0);
	}
	else
	{
		DrawMapSelectFrame(EDITBASEX-2, EDITBASEY-2, EDITSELECTOFF, 0);
		DrawMapSelectFrame(MAPBASEX-2, MAPBASEY-2, MAPSELECTOFF, 1);
		return(1);
	}
}
/****************************************************************************************/
/****************************************************************************************/
void PaintMaps(void)
{
	int i, j, X, Y;
	char v;
	unsigned int FileOff;
	for (j=0; j<LINELENGTH; j++)
		for (i=0; i<LINELENGTH; i++)
		{
			FileOff = j*LINELENGTH + i;
			X = MAPBASEX + i;
			Y = MAPBASEY + j;
			v = *(MapBuffer+FileOff);
			putpixel(screen, X, Y, Map2Pixel((int) v));
		}
	ReadMapCursor();
	DrawMapCursor();
	DrawEditCursor();
	textprintf_centre_ex(screen, font, InstX, 0, WHITE, -1, "Ultima 4 Map Editor v1.1 by Ram Dragon");
	textprintf_centre_ex(screen, font, InstX, InstY, WHITE, -1, "SPACE=toggle maps   ARROWS=move cursor   ESC=quit");
	textprintf_centre_ex(screen, font, InstX, InstY+text_height(font)+6, WHITE, -1, "1=save 2=undo from last save 3=screen shot");
	textprintf_centre_ex(screen, font, InstX, InstY+2*(text_height(font)+6), WHITE, -1, "a..z,A..Z=paint tile");
}
/****************************************************************************************/
/****************************************************************************************/
void EditScreen(int TileNumber)
{
	unsigned int FileOff;
	int X, Y;
	FileOff = MapX*CURSORSIZE + MapY*CURSORSIZE*LINELENGTH + EditX + EditY*LINELENGTH;
	*(MapBuffer+FileOff) = TileNumber;
	X = EditX*CURSOROFF + EDITBASEX;
	Y = EditY*CURSOROFF + EDITBASEY;
	blit(Tiles[TileNumber], screen, 0, 0, X, Y, TILE_SIZE, TILE_SIZE);
	X = MapX*CURSORSIZE + MAPBASEX + EditX;
	Y = MapY*CURSORSIZE + MAPBASEY + EditY;
	EraseMapCursor();
	putpixel(screen, X, Y, Map2Pixel(TileNumber));
	DrawMapCursor();
}
/****************************************************************************************/
/****************************************************************************************/
void PaintEditTiles(void)
{
	int i, X, Y;
	for (i=0; i<10; i++)
	{
		X = i*EDIT_TILE_INC + EDITTILEX;
		Y = EDITTILEY;
		textprintf_centre_ex(screen, font, X, Y, WHITE, -1, "%c", TileLetters[i]);
		blit(Tiles[(TileNumberArray[i])], screen, 0, 0, X, Y+text_height(font), TILE_SIZE, TILE_SIZE);
	}
	for (i=0; i<9; i++)
	{
		X = i*EDIT_TILE_INC + EDITTILEX + (EDIT_TILE_INC/2);
		Y = EDITTILEY + text_height(font) + EDIT_TILE_INC;
		textprintf_centre_ex(screen, font, X, Y, WHITE, -1, "%c", TileLetters[i+10]);
		blit(Tiles[(TileNumberArray[i+10])], screen, 0, 0, X, Y+text_height(font), TILE_SIZE, TILE_SIZE);
	}
	for (i=0; i<7; i++)
	{
		X = i*EDIT_TILE_INC + EDITTILEX + EDIT_TILE_INC;
		Y = EDITTILEY + 2*(text_height(font) + EDIT_TILE_INC);
		textprintf_centre_ex(screen, font, X, Y, WHITE, -1, "%c", TileLetters[i+19]);
		blit(Tiles[(TileNumberArray[i+19])], screen, 0, 0, X, Y+text_height(font), TILE_SIZE, TILE_SIZE);
	}
	for (i=0; i<10; i++)
	{
		X = i*EDIT_TILE_INC + EDITTILEX + 292;
		Y = EDITTILEY;
		textprintf_centre_ex(screen, font, X, Y, WHITE, -1, "%c", TileLetters[i+26]);
		blit(Tiles[(TileNumberArray[i+26])], screen, 0, 0, X, Y+text_height(font), TILE_SIZE, TILE_SIZE);
	}
	for (i=0; i<9; i++)
	{
		X = i*EDIT_TILE_INC + EDITTILEX + (EDIT_TILE_INC/2) + 292;
		Y = EDITTILEY + (text_height(font) + EDIT_TILE_INC);
		textprintf_centre_ex(screen, font, X, Y, WHITE, -1, "%c", TileLetters[i+36]);
		blit(Tiles[(TileNumberArray[i+36])], screen, 0, 0, X, Y+text_height(font), TILE_SIZE, TILE_SIZE);
	}
	for (i=0; i<7; i++)
	{
		X = i*EDIT_TILE_INC + EDITTILEX + EDIT_TILE_INC + 292;
		Y = EDITTILEY + 2*(text_height(font) + EDIT_TILE_INC);
		textprintf_centre_ex(screen, font, X, Y, WHITE, -1, "%c", TileLetters[i+45]);
		blit(Tiles[(TileNumberArray[i+45])], screen, 0, 0, X, Y+text_height(font), TILE_SIZE, TILE_SIZE);
	}
}
/****************************************************************************************/
/****************************************************************************************/
int Map2Pixel(int TileNum)
{
	if (TileNum < 17) return(TilePixelColour[TileNum]); // 0-16
	else if (TileNum > 21 && TileNum < 31) return(TilePixelColour[TileNum-5]); // 22-30 - 5 = 17-25
	else if (TileNum > 47 && TileNum < 56) return(TilePixelColour[TileNum-22]); // 48-55 - 22 = 26-33
	else if (TileNum > 56 && TileNum < 64) return(TilePixelColour[TileNum-23]); // 57-63 - 23 = 34-40
	else if (TileNum > 67 && TileNum < 77) return(TilePixelColour[TileNum-27]); // 68-76 - 27 = 41-49
	else if (TileNum > 125 && TileNum < 127) return(TilePixelColour[TileNum-76]); // 126-127 - 76 = 50-51
	else return(0);
}
/****************************************************************************************/
/****************************************************************************************/
void Map2U4File(char *U4File, char *MapFile)
{
	unsigned long i, j, k, l, MapFileOffset, U4FileOffset;
	for (l=0; l<8; l++)
		for (k=0; k<8; k++)
		{
			MapFileOffset = k*32+l*8192;
			U4FileOffset = k*1024+l*8192;
			for (j=0; j<32; j++)
				for (i=0; i<32; i++)
					*(U4File+U4FileOffset+i+(j*32)) = *(MapFile+MapFileOffset+i+(j*256));
		}
}
/****************************************************************************************/
/****************************************************************************************/
int SaveMap(void)
{
	FILE *fp;
	unsigned long i;
	Map2U4File(Map, MapBuffer);
	if ((fp = fopen("World.map", "wb")) == NULL)
	{
		perror("error open: ");
		return(0);
	}
	if (fwrite(Map, 1, MAP_FILE_SIZE, fp) != MAP_FILE_SIZE)
	{
		perror("error write: ");
		return(0);
	}
	for (i=0; i<65536; i++) *(Map+i) = *(MapBuffer+i);
	fclose(fp);
	return(1);
}
/****************************************************************************************/
/****************************************************************************************/
void RestoreMap(void)
{
	unsigned long i;
	for (i=0; i<65536; i++) *(MapBuffer+i) = *(Map+i);
	PaintMaps();
}
